# Copied from https://github.com/numpy/numpy/pull/9211
#
# Copyright (c) 2005-2017, NumPy Developers.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
#     * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
# 
#     * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.
# 
#     * Neither the name of the NumPy Developers nor the names of any
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np
import numpy._core.numeric as _nx


def _ureduce(a, func, **kwargs):
    a = np.asanyarray(a)
    axis = kwargs.get('axis', None)
    if axis is not None:
        keepdim = list(a.shape)
        nd = a.ndim
        axis = _nx.normalize_axis_tuple(axis, nd)

        for ax in axis:
            keepdim[ax] = 1

        if len(axis) == 1:
            kwargs['axis'] = axis[0]
        else:
            keep = set(range(nd)) - set(axis)
            nkeep = len(keep)
            # swap axis that should not be reduced to front
            for i, s in enumerate(sorted(keep)):
                a = a.swapaxes(i, s)
            # merge reduced axis
            a = a.reshape(a.shape[:nkeep] + (-1,))
            kwargs['axis'] = -1
        keepdim = tuple(keepdim)
    else:
        keepdim = (1,) * a.ndim

    r = func(a, **kwargs)
    return r, keepdim


def percentile(a, q, axis=None, weights=None, out=None,
               overwrite_input=False, interpolation='linear', keepdims=False):
    q = np.true_divide(q, 100.0)  # handles the asarray for us too
    if not _quantile_is_valid(q):
        raise ValueError("Percentiles must be in the range [0, 100]")
    return _quantile_unchecked(
        a, q, axis, weights, out, overwrite_input, interpolation, keepdims)


def quantile(a, q, axis=None, weights=None, out=None,
             overwrite_input=False, interpolation='linear', keepdims=False):
    q = np.asanyarray(q)
    if not _quantile_is_valid(q):
        raise ValueError("Quantiles must be in the range [0, 1]")
    return _quantile_unchecked(
        a, q, axis, weights, out, overwrite_input, interpolation, keepdims)


def _quantile_unchecked(a, q, axis=None, weights=None, out=None,
                        overwrite_input=False, interpolation='linear',
                        keepdims=False):
    """Assumes that q is in [0, 1], and is an ndarray"""
    wgt = _validate_weights(a, q, axis, weights)

    r, k = _ureduce(a, func=_quantile_ureduce_func, q=q, axis=axis,
                    weights=wgt, out=out,
                    overwrite_input=overwrite_input,
                    interpolation=interpolation)
    if keepdims:
        return r.reshape(q.shape + k)
    else:
        return r


def _quantile_is_valid(q):
    # avoid expensive reductions, relevant for arrays with < O(1000) elements
    if q.ndim == 1 and q.size < 10:
        for i in range(q.size):
            if q[i] < 0.0 or q[i] > 1.0:
                return False
    else:
        # faster than any()
        if np.count_nonzero(q < 0.0) or np.count_nonzero(q > 1.0):
            return False
    return True


def _validate_weights(a, q, axis, weights):
    if weights is None:
        wgt = None
    else:
        a = np.asanyarray(a)
        wgt = np.asanyarray(weights)

        if issubclass(a.dtype.type, (np.integer, np.bool_)):
            result_dtype = np.result_type(a.dtype, wgt.dtype, 'f8')
        else:
            result_dtype = np.result_type(a.dtype, wgt.dtype)

        broadcastable = False
        if a.ndim == wgt.ndim and a.shape != wgt.shape:
            broadcastable = all([a_dim == w_dim or w_dim == 1
                                 for a_dim, w_dim
                                 in zip(a.shape, wgt.shape)])

        if a.shape != wgt.shape and not broadcastable:
            if axis is None:
                raise TypeError(
                    "Axis must be specified when shapes of a and weights "
                    "differ and not broadcastable.")
            if wgt.ndim != 1:
                raise TypeError(
                    "1D weights expected when shapes of a and weights differ "
                    " and not broadcastable.")
            if wgt.shape[0] != a.shape[axis]:
                raise ValueError(
                    "Length of weights not compatible with specified axis.")
            if not np.issubdtype(wgt.dtype, np.number):
                raise ValueError("All weight entries must be numeric.")

            nan_ws = wgt[np.isnan(wgt)]
            if nan_ws.size > 0:
                raise ValueError("No weight can be NaN.")

            negative_ws = wgt[wgt < 0]
            if negative_ws.size > 0:
                raise ValueError("Negative weight not allowed.")

            # setup wgt to broadcast along axis
            wgt = np.broadcast_to(wgt, (a.ndim-1)*(1,) + wgt.shape)
            wgt = wgt.swapaxes(-1, axis)
        else:  # same shape, or at least broadcastable
            if axis is None:
                axis = tuple(range(a.ndim))

        scl = wgt.sum(axis=axis, dtype=result_dtype)
        if np.any(scl == 0.0):
            raise ZeroDivisionError(
                "Weights sum to zero, can't be normalized")
        # Obtain a weights array of the same shape as reduced a
        wgt = np.broadcast_to(wgt, a.shape)
        wgt, _ = _ureduce(wgt, func=lambda x, **kwargs: x, axis=axis)

    return wgt


def _quantile_ureduce_func(a, q, axis=None, weights=None, out=None,
                           overwrite_input=False, interpolation='linear',
                           keepdims=False):
    a = np.asarray(a)

    if q.ndim == 0:
        # Do not allow 0-d arrays because following code fails for scalar
        zerod = True
        q = q[None]
    else:
        zerod = False

    if a.size == 1:  # all quantiles point to the same value
        return np.repeat(a, q.size)

    # prepare a for partioning
    if overwrite_input:
        if axis is None:
            ap = a.ravel()
        else:
            ap = a
    else:
        if axis is None:
            ap = a.flatten()
        else:
            ap = a.copy()

    if axis is None:
        axis = 0

    # move axis to -1 for np.vectorize() operations.  Will move back later.
    ap = np.swapaxes(ap, axis, -1)

    if weights is None:
        Nx = ap.shape[-1]
        indices = q * (Nx - 1)
    else:
        # need a copy of weights for later array assignment
        weights = np.swapaxes(weights.astype('f8'), axis, -1)
        weights[weights < 0.] = 0.  # negative weights are treated as 0
        # values with weight=0 are assigned minimum value and later moved left
        abs_min = np.amin(ap)
        ap[weights == 0.] = abs_min - 1.

        def _sort_by_index(vector, vec_indices):
            return vector[vec_indices]
        # this func vectorizes sort along axis
        arraysort = np.vectorize(_sort_by_index, signature='(i),(i)->(i)')

        ind_sorted = np.argsort(ap, axis=-1)  # sort values long axis
        ap_sorted = arraysort(ap, ind_sorted).astype('f8')

        n = np.isnan(ap_sorted[..., -1:])
        if n.ndim > 1:
            n = np.swapaxes(n, axis, -1)

        ws_sorted = arraysort(weights, ind_sorted).astype('f8')
        ws_sorted[np.isnan(ap_sorted)] = 0.  # neglect nans from calculation
        nonzero_w_inds = ws_sorted > 0.

        cum_w = ws_sorted.cumsum(axis=-1)
        cum_w_max = cum_w.max(axis=-1)

        # some manipulation to get lower/upper percentage bounds
        normalized_w_upper = (cum_w.T / cum_w_max.T).T
        prior_cum_w = np.roll(normalized_w_upper, 1, axis=-1)
        prior_cum_w[..., 0] = 0.

        w_slice = ws_sorted  # .copy()
        # in case any input weight is less than 1, we renormalize by min
        if True in (ws_sorted[nonzero_w_inds] < 1.0):
            ws_sorted[nonzero_w_inds] =\
                ws_sorted[nonzero_w_inds] / ws_sorted[nonzero_w_inds].min()

        w_slice[nonzero_w_inds] = ((normalized_w_upper[nonzero_w_inds] -
                                    prior_cum_w[nonzero_w_inds]) /
                                   ws_sorted[nonzero_w_inds])

        w_slice = np.roll(w_slice, -1, axis=-1)
        # now create the lower percentage bound
        normalized_w_lower = np.roll(normalized_w_upper + w_slice, 1, axis=-1)
        normalized_w_lower[..., 0] = 0.0

        # now we subtract by left-most w_slice value
        new_w_upper = (normalized_w_upper.T - w_slice[..., 0].T).T
        new_w_upper[new_w_upper < 0.0] = 0.0
        new_w_lower = (normalized_w_lower.T - w_slice[..., 0].T).T
        new_w_lower[new_w_lower < 0.0] = 0.0
        new_w_lower[..., 0] = 0.0

        # renormalize by right-most bound
        normalized_w_upper = (new_w_upper.T / new_w_upper[..., -1].T).T
        normalized_w_lower = (new_w_lower.T / new_w_upper[..., -1].T).T

        # combine and resort
        cum_w_bands = np.concatenate([normalized_w_upper, normalized_w_lower],
                                     axis=-1)
        inds_resort = np.argsort(cum_w_bands, axis=-1)
        cum_w_bands = arraysort(cum_w_bands, inds_resort)

        ap = np.concatenate([ap_sorted, ap_sorted], axis=-1)
        ap = arraysort(ap, inds_resort)

        # interpolate
        Nx = ap.shape[-1]
        indices_hard = np.arange(Nx)
        vec_interp_func = np.vectorize(np.interp, signature='(n),(m),(m)->(n)')
        indices = vec_interp_func(q, cum_w_bands, indices_hard)

    if interpolation == 'lower':
        indices = np.floor(indices).astype(np.intp)
    elif interpolation == 'higher':
        indices = np.ceil(indices).astype(np.intp)
    elif interpolation == 'midpoint':
        indices = 0.5 * (np.floor(indices) + np.ceil(indices))
    elif interpolation == 'nearest':
        indices = np.around(indices).astype(np.intp)
    elif interpolation == 'linear':
        pass  # keep index as fraction and interpolate
    else:
        raise ValueError(
            "interpolation can only be 'linear', 'lower' 'higher', "
            "'midpoint', or 'nearest'")

    inexact = np.issubdtype(a.dtype, np.inexact)

    if indices.dtype == np.intp:

        if weights is None:
            if inexact:
                indices = np.concatenate((indices, [-1]))  # to move nan's to end
            ap.partition(indices, axis=-1)
            n = np.isnan(ap[..., -1:])
            if inexact:
                indices = indices[:-1]
            if n.ndim > 1:
                n = np.swapaxes(n, axis, -1)

        r = np.take(ap, indices, axis=-1)

        if r.ndim > 1:
            r = np.swapaxes(r, axis, -1)  # move the axis back

        r = np.moveaxis(r, axis, 0)

        if zerod:
            r = r.squeeze(0)

        if out is not None:
            r = np.add(r, 0, out=out)

    else:  # weight the points above and below the indices
        indices_below = np.floor(indices).astype(np.intp)
        indices_above = indices_below + 1
        indices_above[indices_above > Nx - 1] = Nx - 1

        if weights is None:
            if inexact:
                indices_above = np.concatenate((indices_above, [-1]))
            ap.partition(np.concatenate((indices_below, indices_above)), axis=-1)
            n = np.isnan(ap[..., -1:])
            if inexact:
                indices_above = indices_above[:-1]
            if n.ndim > 1:
                n = np.swapaxes(n, axis, -1)

        weights_above = indices - indices_below
        weights_below = 1.0 - weights_above

        def _take1d(vec, inds, wts):
            return np.take(vec, inds) * wts

        vec_take = np.vectorize(_take1d, signature='(n),(m),(m)->(m)')

        x1 = vec_take(ap, indices_below, weights_below)
        x2 = vec_take(ap, indices_above, weights_above)

        if x1.ndim > 1:  # move the axis back
            x1 = np.swapaxes(x1, axis, -1)
            x2 = np.swapaxes(x2, axis, -1)

        x1 = np.moveaxis(x1, axis, 0)
        x2 = np.moveaxis(x2, axis, 0)

        if zerod:
            x1 = x1.squeeze(0)
            x2 = x2.squeeze(0)

        if out is not None:
            r = np.add(x1, x2, out=out)
        else:
            r = np.add(x1, x2)

    if np.any(n):
        warnings.warn("Invalid value encountered in percentile",
                      RuntimeWarning, stacklevel=3)
        if zerod:
            if ap.ndim == 1:
                if out is not None:
                    out[...] = a.dtype.type(np.nan)
                    r = out
                else:
                    r = a.dtype.type(np.nan)
            else:
                r[n.squeeze(axis)] = a.dtype.type(np.nan)
        else:
            if r.ndim == 1:
                r[:] = a.dtype.type(np.nan)
            else:
                r[..., n.squeeze(axis)] = a.dtype.type(np.nan)

    return r



#
# Copyright (C) 2017-2014  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Generate lookup tables for cosmological priors.

This is a lookup table for the BAYESTAR uniform-in-comoving-volume prior. The
BAYESTAR posterior distribution has a builtin prior that is uniform in naively
Euclidean luminosity distance space (proportional to DL^2). When we want to
switch to a prior that is uniform in comoving volume, we have to multiply by
the factor dVC/dVL, the differential comoving volume per unit Euclidean
luminosity volume.

The lookup table is in terms of the *natural logarithm* of the luminosity
distance and provides the *natural logarithm* of dVC/dVL. It consists of the
following constants:

    dVC_dVL_tmin, dVC_dVL_tmax, dVC_dVL_dt
        Minimum, maximum, and step size for a uniform grid in t = log(DL),
        such that t[i] = dVC_dVL_tmin + i * dVC_dVL_dt.

    dVC_dVL_data
        Tabulated values of y = f(t) = log(dVC/dVL), such that y[i] = f(t[i]).

    dVC_dVL_high_z_slope, dVC_dVL_high_z_intercept
        Linear fit for extrapolating f(t) for t>dVC_dVL_tmax.
"""
import argparse
import os
import astropy.cosmology
import astropy.units as u
import numpy as np

from ligo.skymap.postprocess.cosmology import z_for_DL, dVC_dVL_for_DL
from ligo.skymap.util.math import derivative

parser = argparse.ArgumentParser()
parser.add_argument(
    'cosmology', choices=astropy.cosmology.parameters.available,
    default='Planck15', nargs='?', help='Cosmological model')
args = parser.parse_args()

cosmo = getattr(astropy.cosmology, args.cosmology)

DL = np.logspace(0, 6, 32)
log_DL = np.log(DL)
dVC_dVL = dVC_dVL_for_DL(DL)
log_dVC_dVL = np.log(dVC_dVL)


def func(x):
    return np.log(dVC_dVL_for_DL(np.exp(x)))


high_z_x0 = np.log(1e6)
high_z_y0 = func(high_z_x0)
high_z_slope = derivative(func, high_z_x0)
high_z_intercept = high_z_y0 - high_z_slope * high_z_x0

filename = os.path.basename(__file__)
print('/* DO NOT EDIT. Automatically generated by', filename, '*/')
print('static const double dVC_dVL_data[] = {')
print(*('\t{:+.8e}'.format(c) for c in log_dVC_dVL), sep=',\n')
print('};')
print('static const double dVC_dVL_tmin = {:.15f};'.format(log_DL[0]))
print('static const double dVC_dVL_tmax = {:.15f};'.format(log_DL[-1]))
print('static const double dVC_dVL_dt = {:.15f};'.format(
      np.diff(log_DL)[0]))
print('static const double dVC_dVL_high_z_slope = {:.15f};'.format(
      high_z_slope))
print('static const double dVC_dVL_high_z_intercept = {:.15f};'.format(
      high_z_intercept))


def exact_dVC_dVL(DL):
    """An alternate expression of dVC_dVL_for_DL."""
    z = z_for_DL(DL)
    DL = DL * u.Mpc
    DH = cosmo.hubble_distance
    DC = cosmo.comoving_distance(z)
    DM = cosmo.comoving_transverse_distance(z)
    dVC_dz = cosmo.differential_comoving_volume(z)

    Ok0 = cosmo.Ok0
    if Ok0 == 0.0:
        dDM_dDC = 1.0
    elif Ok0 > 0.0:
        dDM_dDC = np.cosh(np.sqrt(Ok0) * DC / DH)
    else:  # Ok0 < 0.0 or Ok0 is nan
        dDM_dDC = np.cos(np.sqrt(-Ok0) * DC / DH)

    dDC_dz = DH * cosmo.inv_efunc(z)
    dDL_dz = DM + (1 + z) * dDM_dDC * dDC_dz
    dVL_dz = np.square(DL) * dDL_dz / u.sr

    return dVC_dz / dVL_dz


DL = np.logspace(-2, 6, 1000)
dVC_dVL = exact_dVC_dVL(DL)

print('static const double dVC_dVL_test_x[] = {')
print(*('\t{:+.8e}'.format(c) for c in DL), sep=',\n')
print('};')
print('static const double dVC_dVL_test_y[] = {')
print(*('\t{:+.8e}'.format(c) for c in dVC_dVL), sep=',\n')
print('};')



#
# Copyright (C) 2015-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Create a contours for the credible levels of an all-sky probability map.
The input is a HEALPix probability map.
The output is a GeoJSON FeatureCollection (http://geojson.org/).
"""

from . import ArgumentParser, FileType


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        '-o', '--output', metavar='FILE.geojson',
        default='-', type=FileType('w'), help='output file [default: stdout]')
    parser.add_argument(
        '--contour', metavar='PERCENT', type=float, nargs='+', required=True,
        help='plot contour enclosing this percentage of probability mass')
    parser.add_argument(
        '-i', '--interpolate',
        choices='nearest nested bilinear'.split(), default='nearest',
        help='resampling interpolation method')
    parser.add_argument(
        '-s', '--simplify', action='store_true', help='simplify contour paths')
    parser.add_argument(
        '-n', '--nside', metavar='NSIDE', type=int,
        help='optionally resample to the specified resolution '
        ' before generating contours')
    parser.add_argument(
        'input', metavar='INPUT.fits[.gz]', type=FileType('rb'),
        default='-', nargs='?', help='Input FITS file')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        import healpy as hp
        import numpy as np
        import json

        from ..io import fits
        from .. import postprocess

        # Read input file
        prob, _ = fits.read_sky_map(opts.input.name, nest=True)

        # Resample if requested
        if opts.nside is not None and opts.interpolate in (
                'nearest', 'nested'):
            prob = hp.ud_grade(prob, opts.nside, order_in='NESTED', power=-2)
        elif opts.nside is not None and opts.interpolate == 'bilinear':
            prob = postprocess.smooth_ud_grade(prob, opts.nside, nest=True)
        if opts.interpolate == 'nested':
            prob = postprocess.interpolate_nested(prob, nest=True)

        # Find credible levels
        i = np.flipud(np.argsort(prob))
        cumsum = np.cumsum(prob[i])
        cls = np.empty_like(prob)
        cls[i] = cumsum * 100

        # Generate contours
        paths = list(postprocess.contour(
            cls, opts.contour, nest=True, degrees=True,
            simplify=opts.simplify))

        json.dump({
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'properties': {
                        'credible_level': contour
                    },
                    'geometry': {
                        'type': 'MultiLineString',
                        'coordinates': path
                    }
                }
                for contour, path in zip(opts.contour, paths)
            ]
        }, opts.output)



import numpy as np
import pytest

from ... import distance
from ... import io
from . import run_entry_point


@pytest.fixture
def skymap(tmpdir):
    npix = 12
    filename = str(tmpdir / 'skymap.fits')
    distmean = 100.0
    diststd = 20.0
    distmu, distsigma, distnorm = distance.moments_to_parameters(
        distmean, diststd)
    io.write_sky_map(filename,
                     (np.tile(1 / npix, npix),
                      np.tile(distmu, npix),
                      np.tile(distsigma, npix),
                      np.tile(distnorm, npix)),
                     gps_time=1e9, objid='foobar',
                     distmean=distmean, diststd=diststd)
    return filename


@pytest.mark.parametrize('geo', [False, True])
@pytest.mark.parametrize('proj', ['mollweide', 'aitoff', 'globe', 'zoom'])
def test_plot(tmpdir, skymap, geo, proj):
    """Test ligo-skymap-plot."""
    pngfilename = str(tmpdir / 'skymap.png')
    args = ['ligo-skymap-plot', skymap, '-o', pngfilename, '--annotate',
            '--colorbar', '--radec', '320', '45', '--projection', proj]
    if geo:
        args.append('--geo')
    if proj == 'globe':
        args.append('--projection-center')
        args.append('0d 0d' if geo else '0h 0d')
    elif proj == 'zoom':
        args.append('--projection-center')
        args.append('0d 0d' if geo else '0h 0d')
        args.append('--zoom-radius')
        args.append('20deg')
    args.extend(['--contour', '90'])
    run_entry_point(*args)


def test_plot_volume(tmpdir, skymap):
    """Test ligo-skymap-plot-volume (at super-low resolution for speed)."""
    pngfilename = str(tmpdir / 'skymap3d.png')
    args = ['ligo-skymap-plot-volume', skymap, '-o', pngfilename, '--dpi', '9',
            '--annotate', '--contour', '90', '--radecdist', '320', '45', '100']
    run_entry_point(*args)



#
# Copyright (C) 2014-2018  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""RGB data for the "Cylon red" color map.

A print- and screen-friendly color map designed specifically for plotting
LSC/Virgo/KAGRA sky maps. The color map is constructed in CIE Lab space,
following a linear ramp in lightness (the `l` coordinate) and a cubic spline in
color components (the `a` and `b` coordinates).

This particular color map was selected from 20 random realizations of this
construction."""

if __name__ == '__main__':  # pragma: no cover
    from colormath.color_conversions import convert_color
    from colormath.color_objects import LabColor, sRGBColor
    from scipy.interpolate import interp1d
    import numpy as np

    def lab_to_rgb(*args):
        """Convert Lab color to sRGB, with components clipped to (0, 1)."""
        Lab = LabColor(*args)
        sRGB = convert_color(Lab, sRGBColor)
        return np.clip(sRGB.get_value_tuple(), 0, 1)

    L_samples = np.linspace(100, 0, 5)

    a_samples = (
        33.34664938,
        98.09940562,
        84.48361516,
        76.62970841,
        21.43276891)

    b_samples = (
        62.73345997,
        2.09003022,
        37.28252236,
        76.22507582,
        16.24862535)

    L = np.linspace(100, 0, 255)
    a = interp1d(L_samples, a_samples[::-1], 'cubic')(L)
    b = interp1d(L_samples, b_samples[::-1], 'cubic')(L)

    for line in __doc__.splitlines():
        print('#', line)
    for L, a, b in zip(L, a, b):
        print(*lab_to_rgb(L, a, b), sep=',')



#
# Copyright (C) 2014-2020  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Register some extra Matplotlib color maps"""

from importlib.resources import files

from matplotlib import colormaps
from matplotlib import colors
import numpy as np

__all__ = ()


for name in ['cylon']:
    # Read in color map RGB data.
    with files(__package__).joinpath(f'{name}.csv').open() as f:
        data = np.loadtxt(f, delimiter=',')

    # Create color map.
    cmap = colors.LinearSegmentedColormap.from_list(name, data)
    # Assign in module.
    locals().update({name: cmap})
    # Register with Matplotlib.
    colormaps.register(cmap=cmap, force=True)

    # Generate reversed color map.
    name += '_r'
    data = data[::-1]
    cmap = colors.LinearSegmentedColormap.from_list(name, data)
    # Assign in module.
    locals().update({name: cmap})
    # Register with Matplotlib.
    colormaps.register(cmap=cmap, force=True)



#
# Copyright (C) 2019-2023  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Bullet charts for Bayes factors."""

from matplotlib import pyplot as plt
import numpy as np

__all__ = ('plot_bayes_factor',)


def plot_bayes_factor(logb,
                      values=(1, 3, 5),
                      labels=('', 'strong', 'very strong'),
                      xlim=7, title=None, palette='RdYlBu',
                      var_label="B"):
    """Visualize a Bayes factor as a `bullet graph`_.

    Make a bar chart of a log Bayes factor as compared to a set of subjective
    threshold values. By default, use the thresholds from
    Kass & Raftery (1995).

    .. _`bullet graph`: https://en.wikipedia.org/wiki/Bullet_graph

    Parameters
    ----------
    logb : float
        The natural logarithm of the Bayes factor.
    values : list
        A list of floating point values for human-friendly confidence levels.
    labels : list
        A list of string labels for human-friendly confidence levels.
    xlim : float
        Limits of plot (`-xlim` to `+xlim`).
    title : str
        Title for plot.
    palette : str
        Color palette.
    var_label : str
        The variable symbol used in plotting

    Returns
    -------
    fig : Matplotlib figure
    ax : Matplotlib axes

    Examples
    --------

    .. plot::
       :include-source:

        from ligo.skymap.plot.bayes_factor import plot_bayes_factor
        plot_bayes_factor(6.3, title='BAYESTAR is awesome')

    """
    with plt.style.context('seaborn-v0_8-notebook'):
        fig, ax = plt.subplots(figsize=(6, 1.7), tight_layout=True)
        ax.set_xlim(-xlim, xlim)
        ax.set_ylim(-0.5, 0.5)
        ax.set_yticks([])
        ax.set_title(title)
        ax.set_ylabel(r'$\ln\,{}$'.format(var_label), rotation=0,
                      rotation_mode='anchor',
                      ha='right', va='center')

        # Add human-friendly labels
        ticks = (*(-x for x in reversed(values)), 0, *values)
        ticklabels = (
            *(f'{s}\nevidence\nagainst'.strip() for s in reversed(labels)), '',
            *(f'{s}\nevidence\nfor'.strip() for s in labels))
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticklabels)
        plt.setp(ax.get_xticklines(), visible=False)
        plt.setp(ax.get_xticklabels()[:len(ticks) // 2], ha='right')
        plt.setp(ax.get_xticklabels()[len(ticks) // 2:], ha='left')

        # Plot colored bands for confidence thresholds
        fmt = plt.FuncFormatter(lambda x, _: f'{x:+g}'.replace('+0', '0'))
        ax2 = ax.twiny()
        ax2.set_xlim(*ax.get_xlim())
        ax2.set_xticks(ticks)
        ax2.xaxis.set_major_formatter(fmt)
        levels = (-xlim, *ticks, xlim)
        colors = plt.get_cmap(palette)(np.arange(1, len(levels)) / len(levels))
        ax.barh(0, np.diff(levels), 1, levels[:-1],
                linewidth=plt.rcParams['xtick.major.width'],
                color=colors, edgecolor='white')

        # Plot bar for log Bayes factor value
        ax.barh(0, logb, 0.5, color='black',
                linewidth=plt.rcParams['xtick.major.width'],
                edgecolor='white')

        for ax_ in fig.axes:
            ax_.grid(False)
            for spine in ax_.spines.values():
                spine.set_visible(False)

    return fig, ax



from astropy.cosmology import Planck15 as cosmo
from astropy import units as u
import lal
import lalsimulation
import numpy as np
import pytest

from ...util.math import derivative
from ..bayestar_inject import get_decisive_snr, cell_max, GWCosmo


def test_get_decisive_snr():
    snrs = [1.0, 3.0, 2.0, 4.0]
    assert get_decisive_snr(snrs, 1) == 4.0
    assert get_decisive_snr(snrs, 2) == 3.0
    assert get_decisive_snr(snrs, 3) == 2.0
    assert get_decisive_snr(snrs, 4) == 1.0
    with pytest.raises(IndexError):
        get_decisive_snr(snrs, 5)


def get_snr_at_z_lalsimulation(cosmo, z, mass1, mass2, f_low, f_high, psd):
    """Calculate optimal SNR the LALSimulation way."""
    params = lal.CreateDict()
    lalsimulation.SimInspiralWaveformParamsInsertRedshift(params, z)

    # "Signal" waveform with requested inclination angle.
    # Take (arbitrarily) only the plus polarization.
    h, _ = lalsimulation.SimInspiralFD(
        mass1 * lal.MSUN_SI, mass2 * lal.MSUN_SI,
        0, 0, 0, 0, 0, 0, cosmo.comoving_distance(z).to_value(u.m),
        0, 0, 0, 0, 0, 1e-3 * (f_high - f_low), f_low, f_high, f_low,
        params, lalsimulation.IMRPhenomPv2)
    return lalsimulation.MeasureSNRFD(h, psd, f_low, f_high)


@pytest.mark.parametrize('mtotal', [2.8, 10.0, 50.0, 100.0])
@pytest.mark.parametrize('z', [0.001, 0.01, 0.1, 1.0, 2.0])
def test_z_at_snr(mtotal, z):
    gwcosmo = GWCosmo(cosmo)
    f_low = 10
    f_high = 4096
    df = 0.1
    mass1 = mass2 = 0.5 * mtotal

    psd = lal.CreateREAL8FrequencySeries(
        '', 0, f_low, df, lal.DimensionlessUnit, int((f_high - f_low) // df))
    lalsimulation.SimNoisePSDaLIGODesignSensitivityP1200087(psd, f_low)

    snr = get_snr_at_z_lalsimulation(
        gwcosmo.cosmo, z, mass1, mass2, f_low, f_high, psd)
    z_solution = gwcosmo.z_at_snr(
        [psd], 'IMRPhenomPv2', f_low, snr, 1, mass1, mass2, 0, 0)

    assert z_solution == pytest.approx(z, rel=1e-2)


def test_get_max_z():
    gwcosmo = GWCosmo(cosmo)
    f_low = 10
    f_high = 4096
    df = 0.1
    waveform = 'IMRPhenomPv2'
    snr = 8
    m1 = np.asarray([50.0])
    m2 = np.asarray([30.0, 50.0])
    x1 = np.asarray([-1.0, 0.0, 1.0])
    x2 = np.asarray([-1.0, -0.5, 0.5, 1.0])

    psd = lal.CreateREAL8FrequencySeries(
        '', 0, f_low, df, lal.DimensionlessUnit, int((f_high - f_low) // df))
    lalsimulation.SimNoisePSDaLIGODesignSensitivityP1200087(psd, f_low)

    result = gwcosmo.get_max_z(
        [psd], waveform, f_low, snr, 1,
        *np.meshgrid(m1, m2, x1, x2, indexing='ij'))
    # Check that shape matches
    assert result.shape == (1, 2, 3, 4)
    # Spot check some individual cells
    for im1, m1_ in enumerate(m1):
        for im2, m2_ in enumerate(m2):
            for ix1, x1_ in enumerate(x1):
                for ix2, x2_ in enumerate(x2):
                    expected = gwcosmo.z_at_snr(
                        [psd], waveform, f_low, snr, 1, m1_, m2_, x1_, x2_)
                    assert result[im1, im2, ix1, ix2] == expected


def test_sensitive_volume_0():
    gwcosmo = GWCosmo(cosmo)
    assert gwcosmo.sensitive_volume(0) == 0


@pytest.mark.parametrize('z', [0.001, 0.01, 0.1, 1.0, 2.0])
def test_sensitive_volume(z):
    """Test that the sensitive volume has the correct derivative with z."""
    gwcosmo = GWCosmo(cosmo)
    dVC_dz = gwcosmo.cosmo.differential_comoving_volume(z)
    expected = (dVC_dz / (1 + z) * 4 * np.pi * u.sr).to_value(u.Mpc**3)
    actual = derivative(
        gwcosmo.sensitive_volume, x0=z, dx=1e-6 * z).to_value(u.Mpc**3)
    assert expected == pytest.approx(actual)


@pytest.mark.parametrize('z', [0.0, 0.001, 0.01, 0.1, 1.0, 2.0])
def test_sensitive_distance(z):
    gwcosmo = GWCosmo(cosmo)
    expected = gwcosmo.sensitive_volume(z).to_value(u.Mpc**3)
    actual = (
        4/3 * np.pi * gwcosmo.sensitive_distance(z)**3).to_value(u.Mpc**3)
    assert expected == pytest.approx(actual)


def test_cell_max():
    values = np.random.uniform(size=(5, 6, 7))
    maxima = cell_max(values)
    np.testing.assert_array_equal(maxima.shape, np.asarray(values.shape) - 1)
    assert np.all(maxima >= values[+1:, +1:, +1:])
    assert np.all(maxima >= values[+1:, +1:, :-1])
    assert np.all(maxima >= values[+1:, :-1, +1:])
    assert np.all(maxima >= values[+1:, :-1, :-1])
    assert np.all(maxima >= values[:-1, +1:, +1:])
    assert np.all(maxima >= values[:-1, +1:, :-1])
    assert np.all(maxima >= values[:-1, :-1, +1:])
    assert np.all(maxima >= values[:-1, :-1, :-1])



#!/usr/bin/env python
# This file is adapted from the Astropy package template, which is licensed
# under a 3-clause BSD style license - see licenses/TEMPLATE_LICENSE.rst

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the setup.cfg file.

import os
import sys

from setuptools import setup


def get_extensions():
    from setuptools import Extension
    import os

    from extension_helpers import add_openmp_flags_if_available, pkg_config
    import numpy as np

    pkg_config_packages = ['gsl']

    orig_kwargs = {
        'sources': [
            'src/bayestar_distance.c',
            'src/bayestar_moc.c',
            'src/bayestar_sky_map.c',
            'src/core.c',
            'src/cubic_interp.c',
            'src/cubic_interp_test.c',
            'src/find_floor.c',
        ],
        'include_dirs': [
            np.get_include()
        ],
        'define_macros': [
            ('GSL_RANGE_CHECK_OFF', None),
            ('HAVE_INLINE', None),
            ('Py_LIMITED_API', 0x030A0000),
            ('NPY_TARGET_VERSION', 'NPY_2_0_API_VERSION'),
            ('NPY_NO_DEPRECATED_API', 'NPY_2_0_API_VERSION'),
        ],
        'extra_compile_args': [
            '-std=gnu11',
            '-fvisibility=hidden'
        ],
    }

    if os.environ.get('LIGO_SKYMAP_USE_SYSTEM_CHEALPIX'):
        pkg_config_packages.append('chealpix')
    else:
        orig_kwargs['include_dirs'].append('cextern/chealpix')
        orig_kwargs['sources'].append('cextern/chealpix/chealpix.c')

    if os.environ.get('LIGO_SKYMAP_USE_ITTNOTIFY'):
        pkg_config_packages.append('ittnotify')
        orig_kwargs['define_macros'].append(('WITH_ITTNOTIFY', 1))

    kwargs = pkg_config(pkg_config_packages, [])
    for key, orig_value in orig_kwargs.items():
        kwargs.setdefault(key, []).extend(orig_value)

    extension = Extension(name='ligo.skymap.core', language='c',
                          py_limited_api=True, **kwargs)

    if not os.environ.get('LIGO_SKYMAP_DISABLE_OPENMP'):
        add_openmp_flags_if_available(extension)

    return [extension]


# First provide helpful messages if contributors try and run legacy commands
# for tests or docs.

TEST_HELP = """
Note: running tests is no longer done using 'python setup.py test'. Instead
you will need to run:

    tox -e test

If you don't already have tox installed, you can install it with:

    pip install tox

If you only want to run part of the test suite, you can also use pytest
directly with::

    pip install -e .[test]
    pytest

For more information, see:

  http://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

if 'test' in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

DOCS_HELP = """
Note: building the documentation is no longer done using
'python setup.py build_docs'. Instead you will need to run:

    tox -e build_docs

If you don't already have tox installed, you can install it with:

    pip install tox

You can also build the documentation with Sphinx directly using::

    pip install -e .[docs]
    cd docs
    make html

For more information, see:

  http://docs.astropy.org/en/latest/install.html#builddocs
"""

if 'build_docs' in sys.argv or 'build_sphinx' in sys.argv:
    print(DOCS_HELP)
    sys.exit(1)

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    version = get_version(root='..', relative_to=__file__)
except Exception:
    version = '{version}'
""".lstrip()


use_scm_version = {'write_to': 'ligo/skymap/version.py',
                   'write_to_template': VERSION_TEMPLATE}

# If we are building under the GitLab CI pipeline and we are building on the
# default branch, then disable the local part of the version
# (+g<short commit hash>) so that we can upload nightly builds to PyPI.
if (
    os.environ.get('CI') == 'true' and
    os.environ.get('CI_COMMIT_BRANCH') == os.environ['CI_DEFAULT_BRANCH']
):
    use_scm_version['local_scheme'] = 'no-local-version'

setup(use_scm_version=use_scm_version, ext_modules=get_extensions())



#
# Copyright (C) 2013-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Create summary plots for sky maps of found injections, optionally binned
cumulatively by false alarm rate or SNR.
"""

import os

from . import ArgumentParser, FileType
from .matplotlib import MatplotlibFigureType


def parser():
    parser = ArgumentParser()
    parser.add_argument('--cumulative', action='store_true')
    parser.add_argument('--normed', action='store_true')
    parser.add_argument(
        '--group-by', choices=('far', 'snr'), metavar='far|snr',
        help='Group plots by false alarm rate (FAR) or ' +
        'signal to noise ratio (SNR)')
    parser.add_argument(
        '--pp-confidence-interval', type=float, metavar='PCT',
        default=95, help='If all inputs files have the same number of '
        'samples, overlay binomial confidence bands for this percentage on '
        'the P--P plot')
    parser.add_argument(
        '--format', default='pdf', help='Matplotlib format')
    parser.add_argument(
        'input', type=FileType('rb'), nargs='+',
        help='Name of input file generated by ligo-skymap-stats')
    parser.add_argument(
        '--output', '-o', default='.', help='output directory')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Imports.
        from astropy.table import Table
        import matplotlib
        matplotlib.use('agg')
        from matplotlib import pyplot as plt
        from matplotlib import rcParams
        import numpy as np
        from tqdm import tqdm
        from .. import plot  # noqa

        # Read in all of the datasets listed as positional command line
        # arguments.
        datasets = [Table.read(file, format='ascii') for file in opts.input]

        # Determine plot colors and labels
        filenames = [file.name for file in opts.input]
        labels = [os.path.splitext(os.path.basename(f))[0] for f in filenames]
        if rcParams['text.usetex']:
            labels = [r'\verb/' + label + '/' for label in labels]
        rcParams['savefig.format'] = opts.format
        metadata = MatplotlibFigureType.get_savefig_metadata(opts.format)

        # Normalize column names
        for dataset in datasets:
            if 'p_value' in dataset.colnames:
                dataset.rename_column('p_value', 'searched_prob')

        if opts.group_by == 'far':

            def key_func(table):
                return -np.log10(table['far'])

            def key_to_dir(key):
                return 'far_1e{}'.format(-key)

            def key_to_title(key):
                return r'$\mathrm{{FAR}} \leq 10^{{{}}}$ Hz'.format(-key)

        elif opts.group_by == 'snr':

            def key_func(table):
                return table['snr']

            def key_to_dir(key):
                return 'snr_{}'.format(key)

            def key_to_title(key):
                return r'$\mathrm{{SNR}} \geq {}$'.format(key)

        else:

            def key_func(table):
                return np.zeros(len(table))

            def key_to_dir(key):
                return '.'

            def key_to_title(key):
                return 'All events'

        if opts.group_by is not None:
            missing = [filename for filename, dataset
                       in zip(filenames, datasets)
                       if opts.group_by not in dataset.colnames]
            if missing:
                raise RuntimeError(
                    'The following files had no "'
                    + opts.group_by + '" column: ' + ' '.join(missing))

        for dataset in datasets:
            dataset['key'] = key_func(dataset)

        if opts.group_by is not None:
            invalid = [filename for filename, dataset
                       in zip(filenames, datasets)
                       if not np.all(np.isfinite(dataset['key']))]
            if invalid:
                raise RuntimeError(
                    'The following files had invalid values in the "'
                    + opts.group_by + '" column: ' + ' '.join(invalid))

        keys = np.concatenate([dataset['key'] for dataset in datasets])

        histlabel = []
        if opts.cumulative:
            histlabel.append('cumulative')
        if opts.normed:
            histlabel.append('fraction')
        else:
            histlabel.append('number')
        histlabel.append('of injections')
        histlabel = ' '.join(histlabel)

        pp_plot_settings = [
            ['', 'searched posterior mass'],
            ['_dist', 'distance CDF at true distance'],
            ['_vol', 'searched volumetric probability']]
        hist_settings = [
            ['searched_area', 'searched_area (deg$^2$)'],
            ['searched_vol', 'searched volume (Mpc$^3$)'],
            ['offset', 'angle from true location and mode of posterior (deg)'],
            ['runtime', 'run time (s)']]

        keys = range(*np.floor([keys.min(), keys.max()+1]).astype(int))
        total = len(keys) * (len(pp_plot_settings) + len(hist_settings))
        with tqdm(total=total) as progress:
            for key in keys:
                filtered = [d[d['key'] >= key] for d in datasets]
                title = key_to_title(key)
                nsamples = {len(d) for d in filtered}
                if len(nsamples) == 1:
                    nsamples, = nsamples
                    title += ' ({} events)'.format(nsamples)
                else:
                    nsamples = None

                subdir = os.path.join(opts.output, key_to_dir(key))
                os.makedirs(subdir, exist_ok=True)

                # Make several different kinds of P-P plots
                for suffix, xlabel in pp_plot_settings:
                    colname = 'searched_prob' + suffix
                    fig = plt.figure(figsize=(6, 6))
                    ax = fig.add_subplot(111, projection='pp_plot')
                    fig.subplots_adjust(bottom=0.15)
                    ax.set_xlabel(xlabel)
                    ax.set_ylabel('cumulative fraction of injections')
                    ax.set_title(title)
                    for d, label in zip(filtered, labels):
                        ax.add_series(d.columns.get(colname, []), label=label)
                    ax.add_diagonal()
                    if nsamples:
                        ax.add_confidence_band(
                            nsamples, 0.01 * opts.pp_confidence_interval)
                    ax.grid()
                    if len(filtered) > 1:
                        ax.legend(loc='lower right')
                    fig.savefig(os.path.join(subdir, colname),
                                metadata=metadata)
                    plt.close()
                    progress.update()

                # Make several different kinds of histograms
                for colname, xlabel in hist_settings:
                    fig = plt.figure(figsize=(6, 4.5))
                    ax = fig.add_subplot(111)
                    fig.subplots_adjust(bottom=0.15)
                    ax.set_xscale('log')
                    ax.set_xlabel(xlabel)
                    ax.set_ylabel(histlabel)
                    ax.set_title(title)
                    values = np.concatenate(
                        [d.columns.get(colname, []) for d in filtered])
                    if len(values) > 0:
                        bins = np.geomspace(np.min(values), np.max(values),
                                            1000 if opts.cumulative else 20)
                        for d, label in zip(filtered, labels):
                            ax.hist(d.columns.get(colname, []),
                                    cumulative=opts.cumulative,
                                    density=opts.normed, histtype='step',
                                    bins=bins, label=label)
                    ax.grid()
                    ax.legend(loc='upper left')
                    fig.savefig(os.path.join(subdir, colname + '_hist'),
                                metadata=metadata)
                    plt.close()
                    progress.update()



# Copyright (C) 2017-2020  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Read events from a GstLal-style SQLite output."""
import os
import sqlite3

from ligo.lw import dbtables

from ...util import sqlite
from .ligolw import LigoLWEventSource

__all__ = ('SQLiteEventSource',)


class SQLiteEventSource(LigoLWEventSource):
    """Read events from LIGO-LW SQLite files.

    Parameters
    ----------
    f : str, file-like object, or `sqlite3.Connection` instance
        The SQLite database.

    Returns
    -------
    `~ligo.skymap.io.events.EventSource`

    """

    def __init__(self, f, *args, **kwargs):
        if isinstance(f, sqlite3.Connection):
            db = f
            filename = sqlite.get_filename(f)
        else:
            if hasattr(f, 'read'):
                filename = f.name
                f.close()
            else:
                filename = f
            db = sqlite.open(filename, 'r')
        super().__init__(dbtables.get_xml(db), *args, **kwargs)
        self._fallbackpath = os.path.dirname(filename) if filename else None


open = SQLiteEventSource



from astropy import units as u
import numpy as np
import pytest

from ..cosmology import cosmo, dVC_dVL_for_DL, dVC_dVL_for_z, dDL_dz_for_z
from ...util.math import derivative

redshift_range = pytest.mark.parametrize('z', np.logspace(-6, 2))


@redshift_range
def test_dDL_dz(z):
    expected = derivative(
        cosmo.luminosity_distance, z, dx=0.001 * z
    ).to_value(u.Mpc)
    result = dDL_dz_for_z(z).to_value(u.Mpc)
    assert expected == pytest.approx(result)


@redshift_range
def test_dVC_dVL(z):
    dVC_dz = cosmo.differential_comoving_volume(z)

    DH = cosmo.hubble_distance
    DL = cosmo.luminosity_distance(z)
    DM = cosmo.comoving_transverse_distance(z)

    dDC_dz = DH * cosmo.inv_efunc(z)  # DC = integral of inv_efunc

    assert cosmo.Ok0 == 0  # Flat universe
    dDM_dz = dDC_dz  # otherwise this expression is more complicated

    dDL_dz = DM + (1 + z) * dDM_dz  # DL = (1 + z) DM

    dVL_dz = DL**2 * dDL_dz * u.sr**-1  # by definition

    expected = (dVC_dz / dVL_dz).to_value(u.dimensionless_unscaled)

    result = dVC_dVL_for_z(z)
    assert expected == pytest.approx(result)

    result = dVC_dVL_for_DL(DL.to_value(u.Mpc))
    assert expected == pytest.approx(result)



#
# Copyright (C) 2015-2024  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Plot a volumetric posterior in three-projection view."""


from . import ArgumentParser, FileType
from .matplotlib import get_figure_parser


def parser():
    parser = ArgumentParser(parents=[get_figure_parser()])
    parser.add_argument(
        '--annotate', default=False, action='store_true',
        help='annotate plot with information about the event')
    parser.add_argument(
        '--max-distance', metavar='Mpc', type=float,
        help='maximum distance of plot in Mpc')
    parser.add_argument(
        '--contour', metavar='PERCENT', type=float, nargs='+',
        help='plot contour enclosing this percentage of'
        ' probability mass')
    parser.add_argument(
        '--radecdist', nargs=3, type=float, action='append', default=[],
        help='right ascension (deg), declination (deg), and distance to mark')
    parser.add_argument(
        '--chain', metavar='CHAIN.hdf5', type=FileType('rb'),
        help='optionally plot a posterior sample chain')
    parser.add_argument(
        '--projection', type=int, choices=list(range(4)), default=0,
        help='Plot one specific projection, or 0 for all projections')
    parser.add_argument(
        'input', metavar='INPUT.fits[.gz]', type=FileType('rb'),
        default='-', nargs='?', help='Input FITS file')
    parser.add_argument(
        '--align-to', metavar='SKYMAP.fits[.gz]', type=FileType('rb'),
        help='Align to the principal axes of this sky map')
    parser.set_defaults(figure_width='3.5', figure_height='3.5')
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Create progress bar.
        from tqdm import tqdm
        progress = tqdm()
        progress.set_description('Starting up')

        # Late imports
        from matplotlib import pyplot as plt
        from matplotlib import gridspec
        from matplotlib import transforms
        from .. import io
        from .. import moc
        from ..plot import marker
        from ..healpix_tree import HEALPIX_MACHINE_ORDER, HEALPIX_MACHINE_NSIDE
        from ..distance import (parameters_to_marginal_moments,
                                principal_axes_moc, volume_render,
                                conditional_pdf, marginal_pdf)
        import healpy as hp
        import numpy as np

        # Read input, determine input resolution.
        progress.set_description('Loading FITS file')
        skymap = io.read_sky_map(opts.input.name, distances=True, moc=True)

        progress.set_description('Preparing projection')

        if opts.align_to is None or opts.input.name == opts.align_to.name:
            skymap2 = skymap
        else:
            skymap2 = io.read_sky_map(
                opts.align_to.name, distances=True, moc=True)
        if opts.max_distance is None:
            mean, std = parameters_to_marginal_moments(
                moc.uniq2pixarea(skymap2['UNIQ']) * skymap2['PROBDENSITY'],
                skymap2['DISTMU'], skymap2['DISTSIGMA'])
            max_distance = mean + 2.5 * std
        else:
            max_distance = opts.max_distance
        rot = np.ascontiguousarray(principal_axes_moc(skymap2))

        if opts.chain:
            chain = io.read_samples(opts.chain.name)
            chain = np.dot(rot.T, (hp.ang2vec(
                0.5 * np.pi - chain['dec'], chain['ra']) *
                np.atleast_2d(chain['dist']).T).T)

        fig = plt.figure()
        n = 1 if opts.projection else 2
        gs = gridspec.GridSpec(
            n, n, left=0.01, right=0.99, bottom=0.01, top=0.99,
            wspace=0.05, hspace=0.05)

        imgwidth = int(opts.dpi * opts.figure_width / n)
        s = np.linspace(-max_distance, max_distance, imgwidth)
        xx, yy = np.meshgrid(s, s)

        truth_marker = marker.reticle(
            inner=0.5 * np.sqrt(2), outer=1.5 * np.sqrt(2), angle=45)

        for iface, (axis0, axis1, (sp0, sp1)) in enumerate((
                (1, 0, [0, 0]),
                (0, 2, [1, 1]),
                (1, 2, [1, 0]),)):

            if opts.projection and opts.projection != iface + 1:
                continue

            progress.set_description(
                'Plotting projection {0}'.format(iface + 1))

            # Marginalize onto the given face
            density = volume_render(
                xx.ravel(), yy.ravel(), max_distance, axis0, axis1, rot,
                skymap).reshape(xx.shape)

            # Plot heat map
            ax = fig.add_subplot(
                gs[0, 0] if opts.projection else gs[sp0, sp1], aspect=1)
            ax.imshow(
                density, origin='lower',
                extent=[-max_distance, max_distance,
                        -max_distance, max_distance],
                cmap=opts.colormap)

            # Add contours if requested
            if opts.contour:
                flattened_density = density.ravel()
                indices = np.argsort(flattened_density)[::-1]
                cumsum = np.empty_like(flattened_density)
                cs = np.cumsum(flattened_density[indices])
                cumsum[indices] = cs / cs[-1] * 100
                cumsum = np.reshape(cumsum, density.shape)
                u, v = np.meshgrid(s, s)
                contourset = ax.contour(
                    u, v, cumsum, levels=opts.contour, linewidths=0.5)

            # Mark locations
            ax._get_lines.get_next_color()  # skip default color
            for ra, dec, dist in opts.radecdist:
                theta = 0.5 * np.pi - np.deg2rad(dec)
                phi = np.deg2rad(ra)
                xyz = np.dot(rot.T, hp.ang2vec(theta, phi) * dist)
                ax.plot(
                    xyz[axis0], xyz[axis1], marker=truth_marker,
                    markerfacecolor='none', markeredgewidth=1)

            # Plot chain
            if opts.chain:
                ax.plot(chain[axis0], chain[axis1], '.k', markersize=0.5)

            # Hide axes ticks
            ax.set_xticks([])
            ax.set_yticks([])

            # Set axis limits
            ax.set_xlim([-max_distance, max_distance])
            ax.set_ylim([-max_distance, max_distance])

            # Mark origin (Earth)
            ax.plot(
                [0], [0], marker=marker.earth, markersize=5,
                markerfacecolor='none', markeredgecolor='black',
                markeredgewidth=0.75)

            if iface == 2:
                ax.invert_xaxis()

        # Add contour labels if contours requested
        if opts.contour:
            ax.clabel(contourset, fmt='%d%%', fontsize=7)

        if not opts.projection:
            # Add scale bar, 1/4 width of the plot
            ax.plot(
                [0.0625, 0.3125], [0.0625, 0.0625],
                color='black', linewidth=1, transform=ax.transAxes)
            ax.text(
                0.0625, 0.0625,
                '{0:d} Mpc'.format(int(np.round(0.5 * max_distance))),
                fontsize=8, transform=ax.transAxes, verticalalignment='bottom')

            # Create marginal distance plot.
            progress.set_description('Plotting distance')
            gs1 = gridspec.GridSpecFromSubplotSpec(5, 5, gs[0, 1])
            ax = fig.add_subplot(gs1[1:-1, 1:-1])

            # Plot marginal distance distribution, integrated over the whole
            # sky.
            d = np.linspace(0, max_distance)
            dp_dd = marginal_pdf(
                d, moc.uniq2pixarea(skymap['UNIQ']) * skymap['PROBDENSITY'],
                skymap['DISTMU'], skymap['DISTSIGMA'], skymap['DISTNORM'])
            ax.fill_between(d, dp_dd, alpha=0.5,
                            color=ax._get_lines.get_next_color())

            if opts.radecdist:
                level, ipix = moc.uniq2nest(skymap['UNIQ'])
                index = ipix << 2 * (HEALPIX_MACHINE_ORDER - level)
                sorter = np.argsort(index)

                # Plot conditional distance distribution at true position
                # and mark true distance.
                for ra, dec, dist in opts.radecdist:
                    match_ipix = hp.ang2pix(HEALPIX_MACHINE_NSIDE, ra, dec,
                                            nest=True, lonlat=True)
                    i = sorter[np.searchsorted(index, match_ipix, side='right',
                                               sorter=sorter) - 1]
                    lines, = ax.plot(
                        [dist], [-0.15], marker=truth_marker,
                        markerfacecolor='none', markeredgewidth=1,
                        clip_on=False,
                        transform=transforms.blended_transform_factory(
                            ax.transData, ax.transAxes))
                    dp_dd = conditional_pdf(
                        d, skymap['DISTMU'][i], skymap['DISTSIGMA'][i],
                        skymap['DISTNORM'][i])
                    ax.fill_between(
                        d, dp_dd, alpha=0.5, color=lines.get_color())
                    ax.axvline(dist, color='black', linewidth=0.5)

            # Scale axes
            ax.set_xticks([0, max_distance])
            ax.set_xticklabels(
                ['0', "{0:d}\nMpc".format(int(np.round(max_distance)))],
                fontsize=9)
            ax.set_yticks([])
            ax.set_xlim(0, max_distance)
            ax.set_ylim(0, ax.get_ylim()[1])

            if opts.annotate:
                text = []
                try:
                    objid = skymap.meta['objid']
                except KeyError:
                    pass
                else:
                    text.append('event ID: {}'.format(objid))
                try:
                    distmean = skymap.meta['distmean']
                    diststd = skymap.meta['diststd']
                except KeyError:
                    pass
                else:
                    text.append('distance: {}±{} Mpc'.format(
                                int(np.round(distmean)),
                                int(np.round(diststd))))
                ax.text(0, 1, '\n'.join(text), transform=ax.transAxes,
                        fontsize=7, ha='left', va='bottom', clip_on=False)

        progress.set_description('Saving')
        opts.output()
